@@ -1,9 +1,3 @@
-/* Special definitions, to process by autoheader.
-   Copyright (C) 1995, 1996 Free Software Foundation.
-
-   Mainly stolen from Ulrich Drepper, <drepper@gnu.ai.mit.edu>
-   gettext-0.10, 1995.  */
-
 @TOP@
 
 /* Define as 1 if you have catgets and don't want to use GNU gettext.  */
@@ -17,6 +11,12 @@
 /* Define if you want to use lookup tables (faster, but uses memory) */
 #undef USE_LOOKUPTABLE
 
+/* Define if you want to use the audiofile library */
+#undef USE_AUDIOFILE
+
+/* Define if you want to use the mad mpeg audio decoder library */
+#undef USE_MAD
+
 @BOTTOM@
 
 #if !HAVE_INT8_T
