@@ -80,32 +80,12 @@
 #include "getopt.h"
 #include "common.h"
 
-#define USE_TEMPFILE 0
-
-#define AMPTODBFS(x) (20 * log10(x))
-#define FRACTODB(x) (20 * log10(x))
-#define DBFSTOAMP(x) pow(10,(x)/20.0)
-#define DBTOFRAC(x) pow(10,(x)/20.0)
-
-extern double signal_max_power(int fd, char *filename,
-			       struct signal_info *psi);
-extern double signal_max_power_stream(FILE *in, char *filename,
-				      struct signal_info *psi);
-extern int apply_gain(int read_fd, int write_fd, char *filename, double gain,
-		      struct signal_info *psi);
+extern double signal_max_power(char *, struct signal_info *);
+extern double signal_max_power_stream(FILE *, char *, struct signal_info *);
+extern int apply_gain(char *fname, double, struct signal_info *);
 
 void compute_levels(struct signal_info *sis, char **fnames, int nfiles);
 double average_levels(struct signal_info *sis, int nfiles, double threshold);
-#if 0
-double amp_to_dBFS(double x);
-double frac_to_dB(double x);
-double dBFS_to_amp(double x);
-double dB_to_frac(double x);
-#endif
-#if USE_TEMPFILE
-int xmkstemp(char *template);
-int xrename(const char *oldpath, const char *newpath);
-#endif
 int strncaseeq(const char *s1, const char *s2, size_t n);
 void *xmalloc(size_t size);
 
@@ -113,16 +93,22 @@ extern char version[];
 char *progname;
 struct progress_struct progress_info;
 
+void
+usage_short()
+{
+  fprintf(stderr, _("Usage: %s [OPTION]... [FILE]...\n"), progname);
+  fprintf(stderr, _("Try `%s --help' for more information.\n"), progname);
+}
 
 void
 usage()
 {
-  fprintf(stderr, _("\
+  printf(_("\
 Usage: %s [OPTION]... [FILE]...\n\
 Normalize volume of multiple WAV files\n\
 \n\
   -a, --amplitude=AMP          normalize the volume to the target amplitude\n\
-                                 AMP [default 0.25 or -12dBFS]\n\
+                                 AMP [default -12dBFS]\n\
   -b, --batch                  batch mode: get average of all levels, and\n\
                                  use one adjustment, based on the average\n\
                                  level, for all files\n\
@@ -141,11 +127,12 @@ Normalize volume of multiple WAV files\n\
       --peak                   adjust by peak level instead of using\n\
                                  loudness analysis\n\
   -q, --quiet                  quiet (decrease verbosity to zero)\n\
-  -t, --average-threshold=TH   when computing average level, ignore any\n\
-                                 levels more than TH decibels from average\n\
-  -T, --adjust-threshold=TH    don't bother applying any adjustment smaller\n\
-                                 than TH decibels\n\
+  -t, --average-threshold=T    when computing average level, ignore any\n\
+                                 levels more than T decibels from average\n\
+  -T, --adjust-threshold=T     don't bother applying any adjustment smaller\n\
+                                 than T decibels\n\
   -v, --verbose                increase verbosity\n\
+  -w, --output-bitwidth=W      force adjusted files to have W-bit samples\n\
 \n\
   -V, --version                display version information and exit\n\
   -h, --help                   display this help and exit\n\
@@ -154,36 +141,48 @@ Report bugs to <cvaill@cs.columbia.edu>.\n"), progname);
 }
 
 enum {
-  OPT_CLIPPING     = 1,
-  OPT_PEAK         = 2,
-  OPT_FRACTIONS    = 3
+  OPT_CLIPPING     = 0x101,
+  OPT_PEAK         = 0x102,
+  OPT_FRACTIONS    = 0x103,
+  OPT_ID3_COMPAT   = 0x104,
+  OPT_ID3_UNSYNC   = 0x105,
+  OPT_NO_PROGRESS  = 0x106,
+  OPT_QUERY        = 0x107,
+  OPT_FRONTEND     = 0x108,
 };
 
-/* arguments */
+/* options */
 int verbose = VERBOSE_PROGRESS;
 int do_print_only = FALSE;
 int do_apply_gain = TRUE;
-double target = 0.25;
+double target = 0.2511886431509580; /* -12dBFS */
 double threshold = -1.0; /* in decibels */
 int do_compute_levels = TRUE;
+int output_bitwidth = 0;
 int gain_in_decibels = FALSE;
 int batch_mode = FALSE;
 int mix_mode = FALSE;
 int use_limiter = TRUE;
 int use_peak = FALSE;
 int use_fractions = FALSE;
+int show_progress = TRUE;
+int do_query = FALSE;
+int frontend = FALSE;
 double lmtr_lvl = 0.5;
 double adjust_thresh = 0.125; /* don't adjust less than this many dB */
+int id3_compat = FALSE;
+int id3_unsync = FALSE;
 
 int
 main(int argc, char *argv[])
 {
-  int fd, fd2, c, i, nfiles;
+  int c, i, nfiles, ret;
   struct signal_info *sis, *psi;
   double level, gain = 1.0, dBdiff;
   char **fnames, *p;
+  char cbuf[32];
   struct stat st;
-  int files_changed = FALSE;
+  int file_needs_adjust = FALSE;
 
   struct option longopts[] = {
     {"help", 0, NULL, 'h'},
@@ -201,12 +200,23 @@ main(int argc, char *argv[])
     {"mix", 0, NULL, 'm'},
     {"compression", 0, NULL, 'c'}, /* deprecate */
     {"limit", 0, NULL, 'c'}, /* deprecate */
+    {"output-bitwidth", 1, NULL, 'w'},
     {"clipping", 0, NULL, OPT_CLIPPING},
     {"peak", 0, NULL, OPT_PEAK},
     {"fractions", 0, NULL, OPT_FRACTIONS},
+    {"id3-compat", 0, NULL, OPT_ID3_COMPAT},
+    {"id3-unsync", 0, NULL, OPT_ID3_UNSYNC},
+    {"no-progress", 0, NULL, OPT_NO_PROGRESS},
+    {"query", 0, NULL, OPT_QUERY},
+    {"frontend", 0, NULL, OPT_FRONTEND},
     {NULL, 0, NULL, 0}
   };
 
+#ifdef __EMX__
+  /* This gives wildcard expansion on Non-POSIX shells with OS/2 */
+  _wildcard(&argc, &argv);
+#endif
+
   /* get program name */
   if ((progname = strrchr(argv[0], '/')) == NULL)
     progname = argv[0];
@@ -222,7 +232,7 @@ main(int argc, char *argv[])
 #endif
 
   /* get args */
-  while ((c = getopt_long(argc, argv, "hVnvqbmcT:l:g:a:t:", longopts, NULL)) != EOF) {
+  while ((c = getopt_long(argc, argv, "hVnvqbmcT:l:g:a:t:w:", longopts, NULL)) != EOF) {
     switch(c) {
     case 'a':
       target = strtod(optarg, &p);
@@ -231,8 +241,8 @@ main(int argc, char *argv[])
       while(isspace(*p))
 	p++;
       if (strncaseeq(p, "db", 2)) {
-
 	/* amplitude given as dBFS */
+
 	if (target > 0) {
 	  target = -target;
 	  fprintf(stderr, _("%s: normalizing to %f dBFS\n"), progname, target);
@@ -270,7 +280,8 @@ main(int argc, char *argv[])
       do_compute_levels = FALSE;
       batch_mode = TRUE;
       if (gain < 0) {
-	usage();
+	fprintf(stderr, _("%s: invalid argument to -g option\n"), progname);
+	usage_short();
 	exit(1);
       }
       break;
@@ -293,7 +304,7 @@ main(int argc, char *argv[])
       /* check if "dB" is given after number */
       while(isspace(*p))
 	p++;
-      fprintf(stderr, _("%s: limiting samples greater than "), progname);
+      /*fprintf(stderr, _("%s: limiting samples greater than "), progname);*/
       if (strncaseeq(p, "db", 2)) {
 	if (lmtr_lvl > 0)
 	  lmtr_lvl = -lmtr_lvl;
@@ -311,8 +322,13 @@ main(int argc, char *argv[])
       adjust_thresh = strtod(optarg, &p);
       if (adjust_thresh < 0)
 	adjust_thresh = -adjust_thresh;
+      /*
       fprintf(stderr, _("%s: ignoring adjustments less than %fdB\n"),
 	      progname, adjust_thresh);
+      */
+      break;
+    case 'w':
+      output_bitwidth = strtol(optarg, NULL, 0);
       break;
     case OPT_CLIPPING:
       use_limiter = FALSE;
@@ -324,6 +340,24 @@ main(int argc, char *argv[])
     case OPT_FRACTIONS:
       use_fractions = TRUE;
       break;
+    case OPT_ID3_COMPAT:
+      id3_compat = TRUE;
+      break;
+    case OPT_ID3_UNSYNC:
+      id3_unsync = TRUE;
+      break;
+    case OPT_NO_PROGRESS:
+      show_progress = FALSE;
+      break;
+    case OPT_QUERY:
+      /*fprintf(stderr, _("%s: Warning: the --query option is deprecated, and may be removed in v1.0\n"),
+	progname);*/
+      do_query = TRUE;
+      break;
+    case OPT_FRONTEND:
+      frontend = TRUE;
+      verbose = VERBOSE_QUIET;
+      break;
     case 'v':
       verbose++;
       break;
@@ -331,16 +365,34 @@ main(int argc, char *argv[])
       verbose = VERBOSE_QUIET;
       break;
     case 'V':
-      fprintf(stderr, "normalize %s\n", version);
+      printf("normalize %s\n", version);
+      printf(_("\
+Copyright (C) 2002 Chris Vaill\n\
+This is free software; see the source for copying conditions.  There is NO\n\
+warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
+"));
+      printf(_("This copy of normalize is compiled with the following libraries:\n"));
+#if USE_MAD
+      printf("  MAD");
+#endif
+#if USE_AUDIOFILE
+      printf("  audiofile");
+#endif
+      printf("\n");
       exit(0);
     case 'h':
       usage();
       exit(0);
     default:
-      usage();
+      usage_short();
       exit(1);
     }
   }
+  if (output_bitwidth < 0 || output_bitwidth > 32) {
+    fprintf(stderr, _("%s: error: output bitwidth must be between 1 and 32\n"),
+ 	    progname);
+    exit(1);
+  }
   if (mix_mode && batch_mode) {
     fprintf(stderr,
 	    _("%s: error: the -m and -b options are mutually exclusive\n"),
@@ -354,8 +406,7 @@ main(int argc, char *argv[])
     exit(1);
   }
   if (optind >= argc) {
-    fprintf(stderr, _("Usage: %s [OPTION]... [FILE]...\n"), progname);
-    fprintf(stderr, _("Try `%s --help' for more information.\n"), progname);
+    usage_short();
     exit(1);
   }
 
@@ -368,6 +419,7 @@ main(int argc, char *argv[])
   fnames = (char **)xmalloc((argc - optind) * sizeof(char *));
   progress_info.file_sizes = (off_t *)xmalloc((argc - optind) * sizeof(off_t));
   for (i = optind; i < argc; i++) {
+#if 0 /* FIXME: read from stdin currently not supported */
     if (strcmp(argv[i], "-") == 0) {
       if (do_apply_gain) {
 	fprintf(stderr, _("%s: Warning: stdin specified on command line, not adjusting files\n"), progname);
@@ -375,13 +427,14 @@ main(int argc, char *argv[])
 	do_print_only = TRUE;
       }
       fnames[nfiles++] = argv[i];
-    } else if (stat(argv[i], &st) == -1) {
+    } else
+#endif
+    if (stat(argv[i], &st) == -1) {
       fprintf(stderr, _("%s: file %s: %s\n"),
 	      progname, argv[i], strerror(errno));
     } else {
-      /* we want the size of the data chunk in kilobytes, so subtract
-         the size of the wav header and divide by size of kb */
-      progress_info.file_sizes[nfiles] = (st.st_size - 36) / 1024;
+      /* we want the size of the file in kilobytes, rounding up */
+      progress_info.file_sizes[nfiles] = (st.st_size + 1023) / 1024;
       /* add the size of the file, in kb */
       progress_info.batch_size += progress_info.file_sizes[nfiles];
       fnames[nfiles] = argv[i];
@@ -395,7 +448,18 @@ main(int argc, char *argv[])
 
   /* allocate space to store levels and peaks */
   sis = (struct signal_info *)xmalloc(nfiles * sizeof(struct signal_info));
+  for (i = 0; i < nfiles; i++) {
+    sis[i].file_size = progress_info.file_sizes[i];
+    sis[i].orig_index = i;
+  }
 
+  if (frontend) {
+    /* frontend mode: print "NUMFILES <number>" */
+    printf("NUMFILES %d\n", nfiles);
+    /* frontend mode: print "FILE <number> <filename>" for each file */
+    for (i = 0; i < nfiles; i++)
+      printf("FILE %d %s\n", i, fnames[i]);
+  }
 
   /*
    * Compute the levels
@@ -427,16 +491,24 @@ main(int argc, char *argv[])
       if (batch_mode)
 	gain = target / level;
 
+      /* frontend mode: print "AVERAGE_LEVEL <level>" */
+      if (frontend)
+	printf("AVERAGE_LEVEL %f\n", AMPTODBFS(level));
+
       if (do_print_only) {
-	if (use_fractions)
-	  printf(_("%-12.6f average level\n"), level);
-	else
-	  printf(_("%-8.4fdBFS average level\n"), AMPTODBFS(level));
+	if (!mix_mode) { /* in mix mode we print everything at the end */
+	  if (use_fractions) {
+	    printf(_("%-12.6f average level\n"), level);
+	  } else {
+	    sprintf(cbuf, "%0.4fdBFS", AMPTODBFS(level));
+	    printf(_("%-12s average level\n"), cbuf);
+	  }
+	}
       } else if (verbose >= VERBOSE_INFO) {
 	if (use_fractions)
-	  fprintf(stderr, _("Average level: %0.4f\n"), level);
+	  printf(_("Average level: %0.4f\n"), level);
 	else
-	  fprintf(stderr, _("Average level: %0.4fdBFS\n"), AMPTODBFS(level));
+	  printf(_("Average level: %0.4fdBFS\n"), AMPTODBFS(level));
       }
     }
 
@@ -444,43 +516,30 @@ main(int argc, char *argv[])
 
 
   /*
+   * FIXME: this comment belongs somewhere else now...
+   *
    * Check if we need to apply the gain --
    *
    *   If a file would be adjusted by an unnoticeable amount, we don't
    *   want to bother doing the adjustment.  The smallest noticeable
    *   difference varies with the absolute intensity and the pitch,
    *   but I don't think it's a stretch to say that a 0.25 dB
-   *   difference is unnoticeable for most pairs of signals.
+   *   difference is unnoticeable for most signals.
    *
    *   By default then, we allow amplitudes that are +/-0.125 dB from
-   *   the target to pass without adjustment (mainly so that
-   *   normalizing files for the second time has no effect).
+   *   the target to pass without adjustment (mainly so that the
+   *   normalize operation is idempotent, i.e. normalizing files for
+   *   the second time has no effect).
    *
    *   Why 0.125 dB?  If we allow amplitudes that are 0.125 dB above
    *   and below the target, the total possible range is 0.25 dB,
    *   which shouldn't be noticeable.
    */
-  if (do_apply_gain && batch_mode) {
+
+  if (batch_mode) {
+    /* if gain_in_decibels, then dBdiff is the original specified gain */
     if (!gain_in_decibels)
       dBdiff = FRACTODB(gain);
-
-    /* the do_compute_levels check makes sure we always apply gain if
-       it was specified with -g */
-    if (do_compute_levels && fabs(dBdiff) < adjust_thresh) {
-      if (verbose >= VERBOSE_PROGRESS)
-	fprintf(stderr,
-		_("Files are already normalized, not adjusting...\n"));
-      do_apply_gain = FALSE;
-    } else if (verbose >= VERBOSE_PROGRESS) {
-      if (!do_compute_levels) { /* if -g */
-	if (gain_in_decibels)
-	  printf(_("Applying adjustment of %fdB...\n"), dBdiff);
-	else
-	  printf(_("Applying adjustment of %f...\n"), gain);
-      } else {
-	printf(_("Applying adjustment of %0.2fdB...\n"), dBdiff);
-      }
-    }
   }
 
 
@@ -488,118 +547,140 @@ main(int argc, char *argv[])
    * Apply the gain
    */
   if (do_apply_gain) {
-#if USE_TEMPFILE
-    struct stat stbuf;
-    char *tmpfile, *p;
-#endif
+
+    /* print adjust message for batch mode */
+    if (batch_mode && verbose >= VERBOSE_PROGRESS) {
+      /* if !do_compute_levels, -g was specified, so we force the adjust */
+      if (do_compute_levels && fabs(dBdiff) < adjust_thresh) {
+	fprintf(stderr, _("Files are already normalized, not adjusting..."));
+      } else {
+	if (!do_compute_levels) { /* if -g */
+	  if (gain_in_decibels)
+	    fprintf(stderr, _("Applying adjustment of %fdB...\n"), dBdiff);
+	  else
+	    fprintf(stderr, _("Applying adjustment of %f...\n"), gain);
+	} else if (do_apply_gain) {
+	  fprintf(stderr, _("Applying adjustment of %0.2fdB...\n"), dBdiff);
+	}
+      }
+    }
 
     progress_info.batch_start = time(NULL);
     progress_info.finished_size = 0;
 
     for (i = 0; i < nfiles; i++) {
 
-      fd = open(fnames[i], O_RDONLY | O_BINARY);
-      if (fd == -1) {
-	fprintf(stderr, _("%s: error opening %s: %s\n"), progname, fnames[i],
-		strerror(errno));
-	continue;
-      }
-
       if (!batch_mode) {
 	if (use_peak)
 	  gain = 1.0 / sis[i].peak;
 	else
 	  gain = target / sis[i].level;
-	dBdiff = FRACTODB(gain);
-
-	/* don't bother applying very small adjustments -- see above */
-	if (fabs(dBdiff) < adjust_thresh) {
-	  if (verbose >= VERBOSE_PROGRESS)
-	    fprintf(stderr, _("%s already normalized, not adjusting...\n"),
-		    fnames[i]);
-	  continue;
-	}
-
-	if (verbose >= VERBOSE_PROGRESS)
-	  fprintf(stderr, _("Applying adjustment of %0.2fdB to %s...\n"),
-		  dBdiff, fnames[i]);
       }
 
-#if USE_TEMPFILE
-      /* Create temporary file name, and open it for writing.  We want
-       * it to be in the same directory (and therefore, in the same
-       * filesystem, for a fast rename). */
-      tmpfile = (char *)xmalloc(strlen(fnames[i]) + 16);
-      strcpy(tmpfile, fnames[i]);
-      if ((p = strrchr(tmpfile, '/')) == NULL)
-	p = tmpfile;
-      else
-	p++;
-      strcpy(p, "_normXXXXXX");
-      fd2 = xmkstemp(tmpfile);
-      if (fd2 == -1) {
-	fprintf(stderr, _("%s: error opening temp file: %s\n"), progname,
-		strerror(errno));
-	close(fd);
-	free(tmpfile);
-	continue;
-      }
-
-      /* preserve original permissions */
-      fstat(fd, &stbuf);
-      fchmod(fd2, stbuf.st_mode);
-#else
-      fd2 = open(fnames[i], O_WRONLY | O_BINARY);
-      if (fd2 == -1) {
-	fprintf(stderr, _("%s: error opening %s: %s\n"), progname, fnames[i],
-		strerror(errno));
-	close(fd);
-	continue;
-      }
-#endif
+      /* frontend mode: print "ADJUSTING <number> <gain>" */
+      if (frontend)
+	printf("ADJUSTING %d %f\n", sis[i].orig_index, FRACTODB(gain));
 
       progress_info.file_start = time(NULL);
       progress_info.on_file = i;
 
       psi = do_compute_levels ? &sis[i] : NULL;
-      if (apply_gain(fd, fd2, fnames[i], gain, psi) == -1) {
+      ret = apply_gain(fnames[i], gain, psi);
+      if (ret == -1) {
 	fprintf(stderr, _("%s: error applying adjustment to %s: %s\n"),
 		progname, fnames[i], strerror(errno));
+      } else {
+	if (ret == 0) {
+	  /* gain was not applied */
+	  if (!batch_mode) {
+	    if (verbose >= VERBOSE_PROGRESS)
+	      fprintf(stderr, _("%s already normalized, not adjusting..."),
+		      fnames[i]);
+	  }
+	} else {
+	  /* gain was applied */
+	  file_needs_adjust = TRUE;
+	}
+	/* frontend mode: print "ADJUSTED <number> 1|0" */
+	if (frontend)
+	  printf("ADJUSTED %d %f\n", sis[i].orig_index, ret);
       }
-      files_changed = TRUE;
 
       progress_info.finished_size += progress_info.file_sizes[i];
 
-      close(fd);
-      close(fd2);
-
-#if USE_TEMPFILE
-      /* move the temporary file back to the original file */
-      if (xrename(tmpfile, fnames[i]) == -1) {
-	fprintf(stderr, _("%s: error moving %s to %s: %s\n"), progname,
-		tmpfile, fnames[i], strerror(errno));
-	exit(1);
-      }
-      free(tmpfile);
-#endif
-
       if (verbose >= VERBOSE_PROGRESS && !batch_mode)
-	fprintf(stderr, "\n");
+	fputc('\n', stderr);
     }
 
     /* we're done with the second progress meter, so go to next line */
     if (verbose >= VERBOSE_PROGRESS && batch_mode)
       fputc('\n', stderr);
 
-  } else if (batch_mode && do_print_only) {
+  } else {
 
-    /* if we're not applying the gain, just print it out, and we're done */
-    if (use_fractions) {
-      printf(_("%-12f volume adjustment\n"), gain);
-    } else {
-      char cbuf[32];
-      sprintf(cbuf, "%fdB", FRACTODB(gain));
-      printf(_("%-12s volume adjustment\n"), cbuf);
+    if (batch_mode && do_print_only) {
+
+      /* if we're not applying the gain, just print it out, and we're done */
+      if (use_fractions) {
+	printf(_("%-12f volume adjustment\n"), gain);
+      } else {
+	sprintf(cbuf, "%fdB", FRACTODB(gain));
+	printf(_("%-12s volume adjustment\n"), cbuf);
+      }
+
+    } else if (mix_mode && do_print_only) {
+
+      /*
+       * In mix mode, we don't have all the information until the end,
+       * so we have to print it all here.
+       */
+      /* clear the progress meter first */
+      if (verbose >= VERBOSE_PROGRESS && show_progress)
+	fprintf(stderr,
+		"\r                                     "
+		"                                     \r");
+      for (i = 0; i < nfiles; i++) {
+	if (use_fractions) {
+	  sprintf(cbuf, "%0.6f", sis[i].level);
+	  printf("%-12s ", cbuf);
+	  sprintf(cbuf, "%0.6f", sis[i].peak);
+	  printf("%-12s ", cbuf);
+	  sprintf(cbuf, "%0.6f", target / sis[i].level);
+	  printf("%-10s ", cbuf);
+	} else {
+	  sprintf(cbuf, "%0.4fdBFS", AMPTODBFS(sis[i].level));
+	  printf("%-12s ", cbuf);
+	  sprintf(cbuf, "%0.4fdBFS", AMPTODBFS(sis[i].peak));
+	  printf("%-12s ", cbuf);
+	  sprintf(cbuf, "%0.4fdB", AMPTODBFS(target / sis[i].level));
+	  printf("%-10s ", cbuf);
+	}
+	printf("%s\n", fnames[i]);
+      }
+      if (use_fractions) {
+	printf(_("%-12.6f average level\n"), level);
+      } else {
+	sprintf(cbuf, "%0.4fdBFS", AMPTODBFS(level));
+	printf(_("%-12s average level\n"), cbuf);
+      }
+
+    } /* end of if (batch_mode && do_print_only) */
+
+    /*
+     * Since we're not applying any gain, we haven't computed
+     * file_needs_adjust yet, so we do it now.
+     */
+    for (i = 0; i < nfiles; i++) {
+      if (use_peak)
+	gain = 1.0 / sis[i].peak;
+      else
+	gain = target / sis[i].level;
+      dBdiff = FRACTODB(gain);
+      
+      if (fabs(dBdiff) >= adjust_thresh) {
+	file_needs_adjust = TRUE;
+	break;
+      }
     }
 
   } /* end of if (do_apply_gain) */
@@ -608,10 +689,23 @@ main(int argc, char *argv[])
   free(progress_info.file_sizes);
   free(fnames);
 
-  /* if the -n option was given but we didn't adjust any files, return
-   * exit status 2 to let scripts know nothing was changed */
-  if (!files_changed && !do_print_only)
-    return 2;
+  /*
+   * frontend mode:
+   *
+   *   print "ADJUST_NEEDED 1" if a file was adjusted, or if the -n
+   *   option was given and a file would need adjustment.
+   *
+   *   print "ADJUST_NEEDED 0" if the -n option was not given and no
+   *   file was adjusted, or if -n was given and no file would need
+   *   adjustment.
+   */
+  if (frontend)
+    printf("ADJUST_NEEDED %d\n", file_needs_adjust ? 1 : 0);
+
+  /* for --query option */
+  /* NOTE: the --query option is broken and obsolete and will go away */
+  if (do_query)
+    return file_needs_adjust ? 0 : 2;
 
   return 0;
 }
@@ -623,9 +717,9 @@ void
 compute_levels(struct signal_info *sis, char **fnames, int nfiles)
 {
   double power;
-  int i, fd;
+  int i;
   char cbuf[32];
-  struct wavfmt fmt = { 1, 2, 44100, 176400, 0, 16 };
+  /*struct wavfmt fmt = { 1, 2, 44100, 176400, 0, 16 };*/
 
   if (verbose >= VERBOSE_PROGRESS) {
     fprintf(stderr, _("Computing levels...\n"));
@@ -643,84 +737,99 @@ compute_levels(struct signal_info *sis, char **fnames, int nfiles)
 
   for (i = 0; i < nfiles; i++) {
 
+    /* frontend mode: print "ANALYZING <number>" for each file index */
+    if (frontend)
+      printf("ANALYZING %d\n", i);
+
     sis[i].level = 0;
 
+#if 0 /* FIXME: reinstate stdin reading */
     if (strcmp(fnames[i], "-") == 0) {
       progress_info.file_start = time(NULL);
       progress_info.on_file = i;
       errno = 0;
 
       /* for a stream, format info is passed through sis[i].fmt */
-      memcpy(&sis[i].fmt, &fmt, sizeof(struct wavfmt));
+      sis[i].channels = 2;
+      sis[i].bits_per_sample = 16;
+      sis[i].samples_per_sec = 44100;
       power = signal_max_power_stream(stdin, NULL, &sis[i]);
       fnames[i] = "STDIN";
 
     } else {
-
-      fd = open(fnames[i], O_RDONLY | O_BINARY);
-      if (fd == -1) {
-	fprintf(stderr, _("%s: error opening %s: %s\n"), progname, fnames[i],
-		strerror(errno));
-	sis[i].level = -1;
-	goto error_update_progress;
-      }
+#endif
 
       progress_info.file_start = time(NULL);
       progress_info.on_file = i;
       errno = 0;
 
-      power = signal_max_power(fd, fnames[i], &sis[i]);
-    }
+      power = signal_max_power(fnames[i], &sis[i]);
 
-    if (power < 0) {
-      fprintf(stderr, _("%s: error reading %s: %s\n"), progname, fnames[i],
-	      strerror(errno));
-      sis[i].level = -1;
-      goto error_close_fd;
+#if 0 /* FIXME */
     }
+#endif
+    if (power < 0) {
+      fprintf(stderr, _("%s: error reading %s"), progname, fnames[i]);
+      if (errno)
+	fprintf(stderr, ": %s\n", strerror(errno));
+      else
+	fprintf(stderr, "\n");
+      sis[i].level = -1;
+      goto error_update_progress;
+    }
+    /* frontend mode: print "LEVEL <number> <level>" for each file index */
+    if (frontend)
+      printf("LEVEL %d %f\n", i, AMPTODBFS(sis[i].level));
     if (power < EPSILON) {
       if (verbose >= VERBOSE_PROGRESS) {
-	fprintf(stderr,
-		"\r                                     "
-		"                                     \r");
+	if (show_progress)
+	  fprintf(stderr,
+		  "\r                                     "
+		  "                                     \r");
 	fprintf(stderr,
 		_("File %s has zero power, ignoring...\n"), fnames[i]);
       }
       sis[i].level = -1;
-      goto error_close_fd;
+      goto error_update_progress;
     }
 
     if (do_print_only) {
 
-      /* clear the progress meter first */
-      if (verbose >= VERBOSE_PROGRESS)
+      /* in mix mode we don't have enough info to print gain yet */
+      if (!mix_mode) {
+
+	/* clear the progress meter first */
+	if (verbose >= VERBOSE_PROGRESS && show_progress)
+	  fprintf(stderr,
+		  "\r                                     "
+		  "                                     \r");
+
+	if (use_fractions) {
+	  sprintf(cbuf, "%0.6f", sis[i].level);
+	  printf("%-12s ", cbuf);
+	  sprintf(cbuf, "%0.6f", sis[i].peak);
+	  printf("%-12s ", cbuf);
+	} else {
+	  sprintf(cbuf, "%0.4fdBFS", AMPTODBFS(sis[i].level));
+	  printf("%-12s ", cbuf);
+	  sprintf(cbuf, "%0.4fdBFS", AMPTODBFS(sis[i].peak));
+	  printf("%-12s ", cbuf);
+	}
+	if (!batch_mode) {
+	  if (use_fractions)
+	    sprintf(cbuf, "%0.6f", target / sis[i].level);
+	  else
+	    sprintf(cbuf, "%0.4fdB", AMPTODBFS(target / sis[i].level));
+	  printf("%-10s ", cbuf);
+	}
+	printf("%s\n", fnames[i]);
+      }
+
+    } else if (verbose >= VERBOSE_INFO) {
+      if (show_progress)
 	fprintf(stderr,
 		"\r                                     "
 		"                                     \r");
-
-      if (use_fractions)
-	sprintf(cbuf, "%0.6f", sis[i].level);
-      else
-	sprintf(cbuf, "%0.4fdBFS", AMPTODBFS(sis[i].level));
-      printf("%-12s ", cbuf);
-      if (use_fractions)
-	sprintf(cbuf, "%0.6f", sis[i].peak);
-      else
-	sprintf(cbuf, "%0.4fdBFS", AMPTODBFS(sis[i].peak));
-      printf("%-12s ", cbuf);
-      if (!batch_mode) {
-	if (use_fractions)
-	  sprintf(cbuf, "%0.6f", target / sis[i].level);
-	else
-	  sprintf(cbuf, "%0.4fdB", AMPTODBFS(target / sis[i].level));
-	printf("%-10s ", cbuf);
-      }
-      printf("%s\n", fnames[i]);
-
-    } else if (verbose >= VERBOSE_INFO) {
-      fprintf(stderr,
-	      "\r                                     "
-	      "                                     \r");
       if (use_fractions)
 	fprintf(stderr, _("Level for %s: %0.4f (%0.4f peak)\n"),
 		fnames[i], sis[i].level, sis[i].peak);
@@ -729,10 +838,6 @@ compute_levels(struct signal_info *sis, char **fnames, int nfiles)
 		fnames[i], AMPTODBFS(sis[i].level), AMPTODBFS(sis[i].peak));
     }
 
-  error_close_fd:
-    if (strcmp(fnames[i], "STDIN") != 0)
-      close(fd);
-
   error_update_progress:
     progress_info.finished_size += progress_info.file_sizes[i];
   }
@@ -791,7 +896,7 @@ average_levels(struct signal_info *sis, int nlevels, double threshold)
     else
       std_dev = sqrt(variance);
     if (verbose >= VERBOSE_INFO)
-      fprintf(stderr, _("Standard deviation is %0.2f dB\n"), std_dev);
+      printf(_("Standard deviation is %0.2f dB\n"), std_dev);
 
     threshold = 2 * std_dev;
   }
@@ -805,21 +910,25 @@ average_levels(struct signal_info *sis, int nlevels, double threshold)
     for (i = 0; i < nlevels; i++) {
 
       /* Find how different from average the i'th file's level is.
-       * The "level" here is actually the signal's maximum amplitude,
+       * The "level" here is the signal's maximum sustained amplitude,
        * from which we can compute the difference in decibels. */
       level_difference = fabs(FRACTODB(mean_level / sis[i].level));
 
       /* mark as bad any level that is > threshold different than the mean */
       if (level_difference > threshold) {
+
+	/* frontend mode: print "AVERAGE_EXCLUDES <number> <difference>" */
+	if (frontend)
+	  printf("AVERAGE_EXCLUDES %d %f\n",
+		 sis[i].orig_index, level_difference);
+
 	if (verbose >= VERBOSE_INFO) {
 	  if (use_fractions) {
-	    fprintf(stderr,
-		_("Throwing out level of %0.4f (different by %0.2fdB)\n"),
-		    sis[i].level, level_difference);
+	    printf(_("Throwing out level of %0.4f (different by %0.2fdB)\n"),
+		   sis[i].level, level_difference);
 	  } else {
-	    fprintf(stderr,
-		_("Throwing out level of %0.4fdBFS (different by %0.2fdB)\n"),
-		    AMPTODBFS(sis[i].level), level_difference);
+	    printf(_("Throwing out level of %0.4fdBFS (different by %0.2fdB)\n"),
+		   AMPTODBFS(sis[i].level), level_difference);
 	  }
 	}
 	badlevels[i] = TRUE;
@@ -848,200 +957,89 @@ average_levels(struct signal_info *sis, int nlevels, double threshold)
   return level;
 }
 
-#define LINE_LENGTH 79
+
 void
 progress_callback(char *prefix, float fraction_completed)
 {
-  char buf[LINE_LENGTH + 32]; /* need +1, but +32 in case of huge ETA's */
+  /* the field lengths used by the sprintf() calls below are carefully
+   * chosen so that buf will never overflow */
+  char buf[128];
   time_t now, time_spent;
   unsigned int file_eta_hr, file_eta_min, file_eta_sec;
   off_t kb_done;
   float batch_fraction;
   unsigned int batch_eta_hr, batch_eta_min, batch_eta_sec;
 
-  if (fraction_completed <= 0.0) {
-    if (progress_info.batch_size == 0) {
-      fprintf(stderr,
-        _(" %s  --%% done, ETA --:--:-- (batch  --%% done, ETA --:--:--)"),
-	      prefix);
-    } else {
-      batch_fraction = (progress_info.finished_size
-			/ (float)progress_info.batch_size);
-      fprintf(stderr,
-        _(" %s  --%% done, ETA --:--:-- (batch %3.0f%% done, ETA --:--:--)"),
-	      prefix, batch_fraction * 100);
-    }
-    fprintf(stderr, "\r");
+  if (!show_progress)
     return;
-  }
 
   now = time(NULL);
+
   if (fraction_completed > 1.0)
     fraction_completed = 1.0;
 
   /* figure out the ETA for this file */
   file_eta_hr = file_eta_sec = file_eta_min = 0;
-  time_spent = now - progress_info.file_start;
-  if (fraction_completed == 0.0)
-    file_eta_sec = 0;
-  else
-    file_eta_sec = floor((float)time_spent / fraction_completed
-			 - (float)time_spent + 0.5);
-  file_eta_min = file_eta_sec / 60;
-  file_eta_sec = file_eta_sec % 60;
-  file_eta_hr = file_eta_min / 60;
-  file_eta_min = file_eta_min % 60;
-  if (file_eta_hr > 99)
-    file_eta_hr = 99;
+  if (fraction_completed > 0.0) {
+    time_spent = now - progress_info.file_start;
+    if (fraction_completed == 0.0)
+      file_eta_sec = 0;
+    else
+      file_eta_sec = (unsigned int)((float)time_spent / fraction_completed
+				    - (float)time_spent + 0.5);
+
+    file_eta_min = file_eta_sec / 60;
+    file_eta_sec = file_eta_sec % 60;
+    file_eta_hr = file_eta_min / 60;
+    file_eta_min = file_eta_min % 60;
+    if (file_eta_hr > 99)
+      file_eta_hr = 99;
+  }
+
 
   /* figure out the ETA for the whole batch */
-  kb_done = progress_info.finished_size
-    + fraction_completed * progress_info.file_sizes[progress_info.on_file];
-  batch_fraction = (float)kb_done / (float)progress_info.batch_size;
   batch_eta_hr = batch_eta_min = batch_eta_sec = 0;
-  time_spent = now - progress_info.batch_start;
-  batch_eta_sec = floor((float)time_spent / batch_fraction
-			- (float)time_spent + 0.5);
+  if (progress_info.batch_size != 0) {
+    kb_done = progress_info.finished_size
+      + fraction_completed * progress_info.file_sizes[progress_info.on_file];
+    batch_fraction = (float)kb_done / (float)progress_info.batch_size;
+    time_spent = now - progress_info.batch_start;
+    batch_eta_sec = (unsigned int)((float)time_spent / batch_fraction
+				   - (float)time_spent + 0.5);
 
-  batch_eta_min = batch_eta_sec / 60;
-  batch_eta_sec = batch_eta_sec % 60;
-  batch_eta_hr = batch_eta_min / 60;
-  batch_eta_min = batch_eta_min % 60;
-  if (batch_eta_hr > 99)
-    batch_eta_hr = 99;
-
-  sprintf(buf, _(" %s %3.0f%% done, ETA %02d:%02d:%02d (batch %3.0f%% done, ETA %02d:%02d:%02d)"),
-	  prefix, fraction_completed * 100,
-	  file_eta_hr, file_eta_min, file_eta_sec,
-	  batch_fraction * 100,
-	  batch_eta_hr, batch_eta_min, batch_eta_sec);
-
-  fprintf(stderr, "%s\r", buf);
-}
-
-
-#if 0
-/*
- * decibel conversion routines
- */
-__inline__ double
-amp_to_dBFS(double x)
-{
-  return 20 * log10(x);
-}
-
-__inline__ double
-frac_to_dB(double x)
-{
-  return 20 * log10(x);
-}
-
-__inline__ double
-dBFS_to_amp(double x)
-{
-  return pow(10, x / 20.0);
-}
-
-__inline__ double
-dB_to_frac(double x)
-{
-  return pow(10, x / 20.0);
-}
-#endif
-
-#if USE_TEMPFILE
-/*
- * This works like the BSD mkstemp, except that we don't unlink the
- * file, since we end up renaming it to something else.
- */
-int
-xmkstemp(char *template)
-{
-  static char sfx[7] = "AAAAAA";
-  char *p;
-  int fd, i, done;
-
-  p = template + strlen(template) - 6;
-  if (strcmp(p, "XXXXXX") != 0) {
-    errno = EINVAL;
-    return -1;
+    batch_eta_min = batch_eta_sec / 60;
+    batch_eta_sec = batch_eta_sec % 60;
+    batch_eta_hr = batch_eta_min / 60;
+    batch_eta_min = batch_eta_min % 60;
+    if (batch_eta_hr > 99)
+      batch_eta_hr = 99;
   }
 
-  do {
-    strcpy(p, sfx);
 
-    /* increment the suffix */
-    done = 0; i = 5;
-    while (!done && i >= 0) {
-      sfx[i]++;
-      if (sfx[i] > 'Z') {
-	sfx[i] = 'A';
-	i--;
-      } else {
-	done = 1;
-      }
-    }
-    if (!done) {
-      errno = EEXIST;
-      return -1;
-    }
-
-    /* attempt to open the file */
-    fd = open(template, O_RDWR | O_CREAT | O_EXCL | O_BINARY, 0600);
-
-  } while (fd == -1 && errno == EEXIST);
-
-  return fd;
-}
-
-
-/*
- * Move the file "oldpath" to "newpath", or copy and delete if they
- * are on different filesystems.
- */
-int
-xrename(const char *oldpath, const char *newpath)
-{
-  if (rename(oldpath, newpath) == -1) {
-    if (errno == EXDEV) {
-      /* files are on different filesystems, so we have to copy */
-      FILE *in, *out;
-      char buf[4096];
-      size_t sz;
-
-      in = fopen(oldpath, "rb");
-      if (in == NULL)
-	return -1;
-      out = fopen(newpath, "wb");
-      if (out == NULL) {
-	fclose(in);
-	return -1;
-      }
-
-      while ((sz = fread(buf, 1, 4096, in)) > 0)
-	fwrite(buf, 1, sz, out);
-
-      if (ferror(in) || ferror(out)) {
-	fclose(in);
-	fclose(out);
-	return -1;
-      }
-      if (fclose(in) == EOF) {
-	fclose(out);
-	return -1;
-      }
-      if (fclose(out) == EOF)
-	return -1;
-      if (unlink(oldpath) == -1)
-	return -1;
+  /* if progress on current file is zero, don't do file ETA */
+  if (fraction_completed <= 0.0) {
+    if (progress_info.batch_size == 0) {
+      /* if we don't have batch info, don't compute batch ETA either */
+      sprintf(buf, _(" %-17s  --%% done, ETA --:--:-- (batch  --%% done, ETA --:--:--)"),
+	      prefix);
     } else {
-      return -1;
+      sprintf(buf, _(" %-17s  --%% done, ETA --:--:-- (batch %3.0f%% done, ETA %02d:%02d:%02d)"),
+	      prefix, batch_fraction * 100,
+	      batch_eta_hr, batch_eta_min, batch_eta_sec);
     }
+
+  } else {
+
+    sprintf(buf, _(" %-17s %3.0f%% done, ETA %02d:%02d:%02d (batch %3.0f%% done, ETA %02d:%02d:%02d)"),
+	    prefix, fraction_completed * 100,
+	    file_eta_hr, file_eta_min, file_eta_sec,
+	    batch_fraction * 100,
+	    batch_eta_hr, batch_eta_min, batch_eta_sec);
+
   }
 
-  return 0;
+  fprintf(stderr, "%s \r", buf);
 }
-#endif
 
 /*
  * Return nonzero if the two strings are equal, ignoring case, up to
